.nolist
include \masm32\include\masm32rt.inc
include \masm32\macros\ucmacros.asm
include \masm32\include\shlwapi.inc
includelib \masm32\lib\shlwapi.lib

.686
.xmm

;
;   CPU architecture structure.
;
STRU_CPU_ARCH               STRUCT
                            wCPU_Arch               WORD ?
                            wReserved               WORD ?
STRU_CPU_ARCH               ENDS

;
;   System Info structure.
;
STRU_SYS_INFO               STRUCT
                                UNION               CPU_TYPE
                                    dOEMID          DWORD ?
                                    CPUArchiture    STRU_CPU_ARCH <>
                                ENDS
                            dPageSize               DWORD ?
                            pMinAppAddress          PDWORD ?
                            pMaxAppAddress          PDWORD ?
                            pActiveCPUMask          PDWORD ?
                            dNumberOfCPUS           DWORD ?
                            dCPUType                DWORD ?
                            dAllocationGramularity  DWORD ?
                            dProcessorLevel         WORD ?
                            dProcessorRevision      WORD ?
STRU_SYS_INFO               ENDS

.data?

ALIGN 16
    dqSave0         OWORD ?
    dqSave1         OWORD ?
    dqSave2         OWORD ?
    dVirtualBase    DWORD ?
    dVirtualSize    DWORD ?
    dSize           DWORD ?
    System_Info     STRU_SYS_INFO <>
    dPageSize       DWORD ?
    dSystemStatus   DWORD ?
    pwSrc           PWORD ?
    pwX             PWORD ?
    pwUse           PWORD ?
ALIGN 16
    wBigBuf         BYTE (65536+16) DUP (?)
    cbDummy         BYTE ?                  ;   Force an unaligned string.
    wCopySrc        WORD 110 DUP (?)
ALIGN 16
    wSrcCopy        BYTE (65536+16) DUP (?)
ALIGN 16
    wPatCopy        BYTE (65536+16) DUP (?)

    WordAlign       PROTO,
                    pwDest:PWORD,
                    pwSrc:PWORD

    KRBWordAlign    PROTO,
                    pwDest:PWORD,
                    pwSrc:PWORD

.code

;-------------------------------------------------------------------------------
;
;   This first test will create two aligned wide character strings (one a single
;   wide character) and will then copy the first string to an unaligned
;   location, then compare the first string and the copy (aligned to unaligned).
;
;-------------------------------------------------------------------------------

start:
	mov pwSrc, uni$("my other brother darryl my other brother darryl my other brother darryl my other brother darryl xxxx")
    mov pwX,   uni$("x")
    INVOKE crt_wcscpy,OFFSET wCopySrc,pwSrc
    or     eax,eax
    jnz    Ok0
mov edi,0
int 3
Ok0:
    INVOKE crt_wcscmp,OFFSET wCopySrc,pwSrc
    or     eax,eax
    jz     Ok1
mov edi,1
int 3
Ok1:
;int 3
;
;   I checked here with an int 3 - the copy was correct, on an unaligned bound.
;

;-------------------------------------------------------------------------------
;
;   This second test will search the unaligned string for a single character.
;
;-------------------------------------------------------------------------------

    mov    eax,pwX
    movzx  ebx,WORD PTR [eax]
    INVOKE crt_wcschr,OFFSET wCopySrc,ebx
    or     eax,eax
    jnz    Ok2
mov edi,2
int 3
Ok2:
;int 3
;
;   I checked here with an int 3 - the first 'x' was found in the unaligned
;   string.
;

;-------------------------------------------------------------------------------
;
;   wcs_ functions are not alignment sensitive! But, non-alignment can cause
;   problems for SSE implementations because aligned loads of SSE regs will get
;   wide characters split across word boundaries in the reg such that pcmpeqw
;   will not work correctly!
;
;   Catch 22! You need aligned SSE reg loads to prevent buffer overrun, but
;   aligned loads will cause character mis-alignment in the SSE reg!
;
;   All hope is not lost, see below.
;
;-------------------------------------------------------------------------------

;
;   Get system info
;
    INVOKE GetSystemInfo,OFFSET System_Info
    mov    eax,System_Info.dPageSize
    mov    dPageSize,eax
    lea    ebx,[eax+65536]
    mov    dVirtualSize,ebx
;
;   Reserve the space required.
;
    INVOKE VirtualAlloc,0,ebx,MEM_RESERVE,PAGE_NOACCESS
    mov    dVirtualBase,eax
    or     eax,eax
    jnz    Ok3
mov edi,3
int 3
Ok3:
;
;   Commit the space required.
;
    INVOKE VirtualAlloc,dVirtualBase,ebx,MEM_COMMIT,PAGE_READWRITE
    or     eax,eax
    jnz    Ok4
mov edi,4
int 3
Ok4:

;-------------------------------------------------------------------------------
;
;   Test for aligned short string at the end of the buffer - this works.
;
;-------------------------------------------------------------------------------

GoodTestAligned:
    mov    ebx,dVirtualBase                 ;   Get the buffer start.
    add    ebx,dVirtualSize                 ;   Get the buffer end.
    sub    ebx,7                            ;   Get the string start (6 NULLS and a NULL).
;
;   This would be the start of the function with string start in ebx.
;
    mov         ecx,ebx                     ;   Save the string start.
    and         ebx,-16                     ;   Force mod 16 alignment.
    mov         edx,0FFFFh                  ;   Get full mask.
    pxor   xmm1,xmm1                        ;   Get 16 nulls.
    sub         ecx,ebx                     ;   Calculate leading garbage characters.
    movdqa xmm0,OWORD PTR [ebx]             ;   Get start of string.
    shl         edx,cl                      ;   Correct the mask for the leading garbage.
    pcmpeqb xmm0,xmm1                       ;   Calculate matches.
    lea         ebx,[ebx+16]                ;   Set for next aligned load.
    pmovmskb ecx,xmm0                       ;   Extract bitmap of matches.
    and         ecx,edx                     ;   Correct the bitmap for garbage the first time.
jnz Ok5
mov edi,5
int 3
Ok5:
    mov         edx,0FFFFh                  ;   Force all remaining checks to be full checks.
    bsf         ecx,ecx                     ;   Get the first match BYTE position.
;
;   ...
;

;-------------------------------------------------------------------------------
;
;   Enter this to test scanning a wide character string starting at an odd BYTE.
;
;   Note: It is always assumed that the wide character strings are properly null
;   terminated and lie within the buffer.
;
;-------------------------------------------------------------------------------

;
;   Start by copying the first 64 bytes of pwSrc to wBigBuf repeatedly until
;   wBigBuf contains 65536 bytes of data, then append a null making this test
;   buffer 1 wide character too long, then set pwUse to the start of this
;   buffer.
;
    mov    esi,pwSrc                        ;   Get the start of the test string.
    mov    edi,OFFSET wBigBuf               ;   Get the start of the buffer.
    mov    pwUse,edi                        ;   Save for testing.
    movdqa xmm0,OWORD PTR[esi]              ;   Fill some xmm regs with the string.
    movdqa xmm1,OWORD PTR[esi+16]
    movdqa xmm2,OWORD PTR[esi+32]
    movdqa xmm3,OWORD PTR[esi+48]
Build:
    movdqa OWORD PTR[edi],xmm0              ;   Save a copy of the string.
    movdqa OWORD PTR[edi+16],xmm1
    movdqa OWORD PTR[edi+32],xmm2
    movdqa OWORD PTR[edi+48],xmm3
    movdqa OWORD PTR[edi+64],xmm0           ;   Save another copy of the string.
    movdqa OWORD PTR[edi+80],xmm1
    movdqa OWORD PTR[edi+96],xmm2
    movdqa OWORD PTR[edi+112],xmm3
    add    edi,128
    cmp    edi,OFFSET wBigBuf+65536
    jb     Build
    mov    WORD PTR [edi],0                 ;   Set a null terminator.

    mov    ecx,65539                        ;   Set the string size plus null plus 1.

;-------------------------------------------------------------------------------
;
;   This next test is checking that an error is returned if the wide character
;   source data is not odd byte aligned for a call to WordAlign.
;
;-------------------------------------------------------------------------------

    mov    esi,OFFSET wSrcCopy              ;   Point to the aligned buffer.
    INVOKE KRBWordAlign,esi,esi             ;   Word align the wide characters in a guaranteed long buffer.
    cmp    eax,1                            ;   Valid return?
    jz     Ok6                              ;   Yes
mov edi,6
int 3
Ok6:

;-------------------------------------------------------------------------------
;
;   This next test is checking that an error is returned if the destination
;   buffer is not 16 byte aligned for a call to WordAlign.
;
;-------------------------------------------------------------------------------

    lea    esi,[esi+1]                      ;   Unalign the pointer to the aligned buffer.
    INVOKE KRBWordAlign,esi,esi             ;   Word align the wide characters in a guaranteed long buffer.
    cmp    eax,2                            ;   Valid return?
    jz     Ok7                              ;   Yes
mov edi,7
int 3
Ok7:

;-------------------------------------------------------------------------------
;
;   This next test is checking that a string that is too long can be detected.
;
;-------------------------------------------------------------------------------

    mov    ebx,dVirtualBase                 ;   Get the buffer start.
    inc    ebx                              ;   Start at odd bound.
    mov    edi,pwUse
    mov    WORD PTR [edi+ecx-3],0           ;   Truncate the source string at string size.
;
;   Copy the string to the end of the VirtualAlloc buffer.
;
    pushad
    INVOKE crt_wcscpy,ebx,edi
    popad

    mov    esi,OFFSET wSrcCopy              ;   Point to the aligned buffer.
    INVOKE KRBWordAlign,esi,ebx             ;   Word align the wide characters in a guaranteed long buffer.
    cmp    eax,3                            ;   Valid return?
    jz     Ok8                              ;   Yes
mov edi,8
int 3
Ok8:

    mov    ecx,65535                        ;   Set the string size plus null plus 1.

;-------------------------------------------------------------------------------
;
;   This test loop is checking that a right justified null terminated wide
;   character string saved at an odd BYTE address at the end of a VirtualAlloc
;   buffer can be properly aligned in a 16 BYTE aligned buffer.
;
;-------------------------------------------------------------------------------

GoodTestUnalignedWide:
    mov    ebx,dVirtualBase                 ;   Get the buffer start.
    add    ebx,dVirtualSize                 ;   Get the buffer end.
    sub    ebx,ecx                          ;   Get the string start.
    mov    edi,pwUse
    mov    WORD PTR [edi+ecx-3],0           ;   Truncate the source string at string size.
;
;   Copy the string to the end of the VirtualAlloc buffer.
;
    pushad
    INVOKE crt_wcscpy,ebx,edi
    popad

    mov    esi,OFFSET wSrcCopy              ;   Point to the aligned buffer.
    INVOKE KRBWordAlign,esi,ebx             ;   Word align the wide characters in a guaranteed long buffer.
    test   eax,0Fh                          ;   Valid return?
    jz     Ok9                              ;   Yes
mov edi,9
int 3
Ok9:
    mov    esi,eax
;
;   Check for a valid copy.
;
    pushad
    INVOKE crt_wcscmp,esi,ebx
    cmp    eax,0
    popad
    jz     Ok10

mov edi,10
int 3
Ok10:
;
;   String done, get the Pattern.
;
DoPat:
    mov    ebx,dVirtualBase                 ;   Get the buffer start.
    add    ebx,dVirtualSize                 ;   Get the buffer end.
    sub    ebx,ecx                          ;   Get the string start.

    pushad
    INVOKE crt_wcscpy,ebx,pwUse
    popad

    mov    edi,OFFSET wPatCopy              ;   Point to the aligned buffer.
    INVOKE KRBWordAlign,edi,ebx             ;   Word align the wide characters in a guaranteed long buffer.
    test   eax,0Fh                          ;   Valid return?
    jz     Ok11                             ;   Yes
mov edi,11
int 3
Ok11:
    mov    edi,eax
;
;   Check for a valid copy.
;
    pushad
    INVOKE crt_wcscmp,edi,ebx
    cmp    eax,0
    popad
    jz     Ok12

mov edi,12
int 3
Ok12:
;
;   esi edi are aligned for compares, or whatever. 
;
IsAligned:
;
;   ...
;
    lea    ecx,[ecx-2]                      ;   Test all string lengths down to 1.
    cmp    ecx,3
    ja     GoodTestUnalignedWide

;-------------------------------------------------------------------------------
;
;   Test for short string at the end of the buffer - this does not work.
;
;   This will fail for either aligned or mis-aligned strings.
;
;   The failure will an "access violation reading..."  for a load of xmm7.
;
;-------------------------------------------------------------------------------

BadTest:
    mov    ebx,dVirtualBase                 ;   Get the buffer start.

    add    ebx,dVirtualSize                 ;   Get the buffer end.
    sub    ebx,7                            ;   Get the string start (6 NULLS and a NULL).
;
;   This would be the start of the function with string start in ebx.
;
    movdqu xmm7,OWORD PTR [ebx]             ;   This will fail right here.
mov edi,13
int 3

;-------------------------------------------------------------------------------
;
;   Exit the program.
;
;-------------------------------------------------------------------------------

	exit

;-------------------------------------------------------------------------------
;
;   Align an unaligned wide character null terminated string in a buffer,
;   checking for nulls, copying it to an aligned buffer.
;
;   Unaligned in this case means the wide chararacter string starts on an odd
;   BYTE bound. The buffer is expected to be mod 16 bounded, and 65536+16 bytes
;   long (maximum supported string length + 16 byte pad to prevent any overrun).
;
;   The only safe unaligned (odd BYTE start) sse load is at a mod 15 BYTE start.
;   Any aligned sse load that contains the start of the string is also valid.
;
;   If the pointer to the first character of a fragment load is a mod 15 bound,
;   then the first character lies split across two 16 byte aligned blocks.
;   Because of this split, the unaligned load will always be valid because part
;   of the first character and a possible null must follow in the next aligned
;   16 byte block. For every mod 15 bound unaligned block loaded, either there
;   are no nulls (the following block must exist), a null is found (no need to
;   look past this block), or the null is split across the block boundary, thus
;   the next block must exist.
;
;   Since this is a valid null terminated string, an aligned load that contains
;   the start of the string must be valid since VirtualAlloc allocated buffers
;   start on page bounds (4 KB) and are modulo page size (4 KB), thus since the
;   string lies in a VirtualAlloc allocated area, the start of the string must
;   be loadable with an aligned sse load that contains the start of the string.
;   Subsequent loads (if a null is not detected) will be set up to load from a
;   mod 15 BYTE start, thus always be valid.
;
;   Returns the pointer to the buffer, or returns a flag if an error is detected
;   (1 if wide character string does not start on a odd BYTE bound, 2 if buffer
;   is not mod 16 bounded, 3 if wide character string is too long). It preserves
;   all regs used (including sse regs used) except the returned eax.
;
;   Test the return least significant 4 bits to check for validity (all 4 bits
;   should be 0 if valid).
;
;-------------------------------------------------------------------------------

WordAlign   PROC USES ebx ecx esi edi,
            pwBuf:PWORD,                    ;   Aligned buffer for copy. 
            pwStr:PWORD                     ;   Unaligned wide string.

    movdqa      OWORD PTR dqSave0,xmm0      ;   Save the xmm regs to be used.
    movdqa      OWORD PTR dqSave1,xmm1
    movdqa      OWORD PTR dqSave2,xmm2
    mov    ebx,pwStr                        ;   Get the unaligned string pointer.
    mov    edi,pwBuf                        ;   Get the aligned buffer pointer.
    mov    eax,1                            ;   Set pwStr not odd error return.
    test   ebx,1                            ;   pwStr valid?
    jz     Exit                             ;   No.
    mov    eax,2                            ;   Set pwBuf not mod 16 error return.
    test   edi,0Fh                          ;   pwBuf valid?
    jnz    Exit                             ;   No.
    mov         esi,ebx                     ;   Save pwStr for DoUnaligned.
    mov    ecx,ebx                          ;   Save the string start.
    and    ebx,-16                          ;   Force mod 16 alignment.
    sub    ecx,ebx                          ;   Calculate leading garbage BYTES.
    pxor        xmm1,xmm1                   ;   Get 16 nulls for DoUnaligned also.
    cmp         ecx,15                      ;   No valid WORDS (only half of one WORD for the xmm reg)?
    jz          Setup                       ;   Yes, just use unaligned loads.
    lea    ecx,[ecx-1]                      ;   Decrement ecx for the alignment shift.
    movdqa      xmm0,OWORD PTR [ebx]        ;   Get aligned data containing the start of the string.
    psrldq      xmm0,1                      ;   WORD align the wide characters left (discard first BYTE).
    movdqa      xmm2,xmm0                   ;   Copy the aligned data.
    pcmpeqw     xmm0,xmm1                   ;   Calculate null matches on remaining 7 WORDS.
    pmovmskb    eax,xmm0                    ;   Extract bitmap (16 bits for 8 WORDS) of null matches.
    shr    ax,cl                            ;   Mask the matches for valids.
;
;   Shift out garbage WORDS preceeding the string.
;
    test   ecx,2                            ;   Test the WORD shift count for 1 (ecx is BYTES here).
    jz     SkipOne                          ;   Not shift 1.
    psrldq      xmm2,2                      ;   Shift the wide characters left 1 WORD.
SkipOne:
    test   ecx,4                            ;   Test the WORD shift count for 2.
    jz     SkipTwo                          ;   Not shift 2.
    psrldq      xmm2,4                      ;   Shift the wide characters left 2 WORDS.
SkipTwo:
    test   ecx,8                            ;   Test the WORD shift count for 4.
    jz     SkipThree                        ;   Not shift 4.
    psrldq      xmm2,8                      ;   Shift the wide characters left 4 WORDS.
SkipThree:
    movdqa      OWORD PTR [edi],xmm2        ;   Save the string head.
    test   eax,eax                          ;   Any valid null matches found?
    jnz    Good                             ;   Yes, go no further.
;
;   Set load and store pointers to skip these saved leading characters.
;
    mov    ebx,14                           ;   Total bytes that could have been in xmm0.
    sub    ebx,ecx                          ;   Total valid bytes that remain.
    lea    esi,[esi+ebx]                    ;   Set address for next unaligned load.
    lea    edi,[edi+ebx]                    ;   Set address for next unaligned store.
;
;   Set up overflow check.
;
Setup:
    mov    eax,pwBuf                        ;   Point to the buffer.
    lea    eax,[eax+65536]                  ;   Set the first invalid address.
;
;   Move unaligned data to aligned buffer.
;
DoUnaligned:
    movdqu      xmm0,OWORD PTR [esi]        ;   Get unaligned data, guaranteed no buffer overrun.
    movdqu      OWORD PTR [edi],xmm0        ;   Build an unaligned buffer, guaranteed no buffer overrun.
    pcmpeqw     xmm0,xmm1                   ;   Calculate null matches.
    pmovmskb    ecx,xmm0                    ;   Extract bitmap of null matches.
    lea    esi,[esi+16]                     ;   Set up for next load.
    lea    edi,[edi+16]                     ;   Set up for next store.
    test   ecx,ecx                          ;   Any nulls?
    jnz    CheckNull                        ;   Yes, check if too big.
    cmp    edi,eax                          ;   Next store valid?
    jb     DoUnaligned                      ;   Yes.
    jmp    Overflow                         ;   No.
;
;   Check if the null is out of buffer.
;
CheckNull:
    bsf    ebx,ecx                          ;   Get the BYTE number of the lowest null.
    lea    edi,[edi+ebx-16]                 ;   Get the address of the null.
    cmp    edi,eax                          ;   Valid store address?
    jae    Overflow                         ;   No.
;
;   Set good return (pointer to buffer - pwBuf).
;
Good:
    mov    eax,pwBuf                        ;   Set good return.
    jmp    Exit
;
;   Set buffer overflow error return.
;
Overflow:
    mov    eax,3
    jmp    Exit
;
;   Exit WordAlign.
;
Exit:
    movdqa      xmm0,OWORD PTR dqSave0      ;   Restore the xmm regs used.
    movdqa      xmm1,OWORD PTR dqSave1
    movdqa      xmm2,OWORD PTR dqSave2
    ret    8
WordAlign   ENDP

;-------------------------------------------------------------------------------
;
;   Align an unaligned wide character null terminated string in a buffer,
;   checking for nulls, copying it to an aligned buffer.
;
;   Unaligned in this case means the wide chararacter string starts on an odd
;   BYTE bound. The buffer is expected to be mod 16 bounded, and 65536+16 bytes
;   long (maximum supported string length + 16 byte pad to prevent any overrun).
;
;   The only safe unaligned (odd BYTE start) sse load is at a mod 15 BYTE start.
;   Any aligned sse load that contains the start of the string is also valid.
;
;   If the pointer to the first character of a fragment load is a mod 15 bound,
;   then the first character lies split across two 16 byte aligned blocks.
;   Because of this split, the unaligned load will always be valid because part
;   of the first character and a possible null must follow in the next aligned
;   16 byte block. For every mod 15 bound unaligned block loaded, either there
;   are no nulls (the following block must exist), a null is found (no need to
;   look past this block), or the null is split across the block boundary, thus
;   the next block must exist.
;
;   Since this is a valid null terminated string, an aligned load that contains
;   the start of the string must be valid since VirtualAlloc allocated buffers
;   start on page bounds (4 KB) and are modulo page size (4 KB), thus since the
;   string lies in a VirtualAlloc allocated area, the start of the string must
;   be loadable with an aligned sse load that contains the start of the string.
;   Subsequent loads (if a null is not detected) will be set up to load from a
;   mod 15 BYTE start, thus always be valid.
;
;   Returns the pointer to the buffer, or returns a flag if an error is detected
;   (1 if wide character string does not start on a odd BYTE bound, 2 if buffer
;   is not mod 16 bounded, 3 if wide character string is too long). It preserves
;   all regs used (including sse regs used) except the returned eax.
;
;   Test the return least significant 4 bits to check for validity (all 4 bits
;   should be 0 if valid).
;
;-------------------------------------------------------------------------------

KRBWordAlign    PROC USES ebx ecx edx esi edi,
                pwBuf:PWORD,                ;   Aligned buffer for copy. 
                pwStr:PWORD                 ;   Unaligned wide string.

    LOCAL   dqSSave[4]:OWORD

    lea         edx,dqSSave + 15
    and         edx,-16
    movdqa      OWORD PTR [edx],xmm0        ;   Save the xmm regs to be used.
    movdqa      OWORD PTR [edx+16],xmm1
    movdqa      OWORD PTR [edx+32],xmm2
    mov    ebx,pwStr                        ;   Get the unaligned string pointer.
    mov    edi,pwBuf                        ;   Get the aligned buffer pointer.
    mov    eax,1                            ;   Set pwStr not odd error return.
    test   ebx,1                            ;   pwStr valid?
    jz     Exit                             ;   No.
    mov    eax,2                            ;   Set pwBuf not mod 16 error return.
    test   edi,0Fh                          ;   pwBuf valid?
    jnz    Exit                             ;   No.
    mov         esi,ebx                     ;   Save pwStr for DoUnaligned.
    mov    ecx,ebx                          ;   Save the string start.
    and    ebx,-16                          ;   Force mod 16 alignment.
    sub    ecx,ebx                          ;   Calculate leading garbage BYTES.
    pxor        xmm1,xmm1                   ;   Get 16 nulls for DoUnaligned also.
    cmp         ecx,15                      ;   No valid WORDS (only half of one WORD for the xmm reg)?
    jz          SetupA                      ;   Yes, just use unaligned loads, aligned stores.
    movdqa      xmm0,OWORD PTR [ebx]        ;   Get aligned data containing the start of the string.
    lea    ecx,[ecx-1]                      ;   Decrement ecx for the alignment shift.
    psrldq      xmm0,1                      ;   WORD align the wide characters left (discard first BYTE).
    mov    ebx,14                           ;   Total bytes that could have been in xmm0.
    movdqa      xmm2,xmm0                   ;   Copy the aligned data.
    pcmpeqw     xmm0,xmm1                   ;   Calculate null matches on remaining 7 WORDS.
    pmovmskb    eax,xmm0                    ;   Extract bitmap (16 bits for 8 WORDS) of null matches.
    shr    ax,cl                            ;   Mask the matches for valids.
;
;   Shift out garbage WORDS preceeding the string.
;
    test   ecx,2                            ;   Test the WORD shift count for 1 (ecx is BYTES here).
    jz     SkipOne                          ;   Not shift 1.
    psrldq      xmm2,2                      ;   Shift the wide characters left 1 WORD.
SkipOne:
    test   ecx,4                            ;   Test the WORD shift count for 2.
    jz     SkipTwo                          ;   Not shift 2.
    psrldq      xmm2,4                      ;   Shift the wide characters left 2 WORDS.
SkipTwo:
    test   ecx,8                            ;   Test the WORD shift count for 4.
    jz     SkipThree                        ;   Not shift 4.
    psrldq      xmm2,8                      ;   Shift the wide characters left 4 WORDS.
SkipThree:
    movdqa      OWORD PTR [edi],xmm2        ;   Save the string head.
    test   eax,eax                          ;   Any valid null matches found?
    jnz    Good                             ;   Yes, go no further.
;
;   Set load and store pointers to skip these saved leading characters.
;
    sub    ebx,ecx                          ;   Total valid bytes that remain.
    lea    esi,[esi+ebx]                    ;   Set address for next unaligned load.
    lea    edi,[edi+ebx]                    ;   Set address for next unaligned store.
;
;   Set up overflow check.
;
SetupU:
    mov    eax,pwBuf                        ;   Point to the buffer.
    lea    eax,[eax+65536]                  ;   Set the first invalid address.
;
;   Move unaligned data to unaligned buffer.
;
DoUnaligned:
    movdqu      xmm0,OWORD PTR [esi]        ;   Get unaligned data, guaranteed no buffer overrun.
    lea    esi,[esi+16]                     ;   Set up for next load.
    movdqu      OWORD PTR [edi],xmm0        ;   Build an unaligned buffer, guaranteed no buffer overrun.
    lea    edi,[edi+16]                     ;   Set up for next store.
    pcmpeqw     xmm0,xmm1                   ;   Calculate null matches.
    pmovmskb    ecx,xmm0                    ;   Extract bitmap of null matches.
    test   ecx,ecx                          ;   Any nulls?
    jnz    CheckNull                        ;   Yes, check if too big.
    cmp    edi,eax                          ;   Next store valid?
    jb     DoUnaligned                      ;   Yes.
    jmp    Overflow                         ;   No.
;
;   Set up overflow check. SetupA and DoAligned is the same as SetupU and
;   DoUnaligned with the exception that the store is aligned for DoAligned.
;
SetupA:
    mov    eax,pwBuf                        ;   Point to the buffer.
    lea    eax,[eax+65536]                  ;   Set the first invalid address.
;
;   Move unaligned data to aligned buffer.
;
DoAligned:
    movdqu      xmm0,OWORD PTR [esi]        ;   Get unaligned data, guaranteed no buffer overrun.
    lea    esi,[esi+16]                     ;   Set up for next load.
    movdqa      OWORD PTR [edi],xmm0        ;   Build an aligned buffer, guaranteed no buffer overrun.
    lea    edi,[edi+16]                     ;   Set up for next store.
    pcmpeqw     xmm0,xmm1                   ;   Calculate null matches.
    pmovmskb    ecx,xmm0                    ;   Extract bitmap of null matches.
    test   ecx,ecx                          ;   Any nulls?
    jnz    CheckNull                        ;   Yes, check if too big.
    cmp    edi,eax                          ;   Next store valid?
    jb     DoAligned                        ;   Yes.
    jmp    Overflow                         ;   No.
;
;   Check if the null is out of buffer.
;
CheckNull:
    bsf    ebx,ecx                          ;   Get the BYTE number of the lowest null.
    lea    edi,[edi+ebx-16]                 ;   Get the address of the null.
    cmp    edi,eax                          ;   Valid store address?
    jae    Overflow                         ;   No.
;
;   Set good return (pointer to buffer - pwBuf).
;
Good:
    mov    eax,pwBuf                        ;   Set good return.
    jmp    Exit
;
;   Set buffer overflow error return.
;
Overflow:
    mov    eax,3
    jmp    Exit
;
;   Exit WordAlign.
;
Exit:
    movdqa      xmm0,OWORD PTR [edx]        ;   Restore the xmm regs used.
    movdqa      xmm1,OWORD PTR [edx+16]
    movdqa      xmm2,OWORD PTR [edx+32]
    ret    8
KRBWordAlign    ENDP

end start

